/* $Id: nfs_cache.c,v 1.38 1997/07/10 16:37:41 gjb Exp $
 * nfs_cache.c
 * (C) 1997 Greg J. Badros -- All Rights Reserved
 * <gjb@cs.washington.edu> 
 * Seattle, WA  USA
 * http://www.cs.washington.edu/homes/gjb
 */

#define NFS_NEED_XDR_TYPES
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>
/*#include <linux/errno.h> */

#include <asm/segment.h>
#include <asm/system.h>

#include "profile.h"
#include "params.h"
#include "nfs_cache.h"

#define NFSDBG_FACILITY		NFSDBG_PAGECACHE

/* pointer to the head of the linked lists of cache data structures */
static struct nfs_cache *nfs_cache_first = NULL;

static
int /* non-zero if found */
nfs_create_new_cache_if_no_reuse(struct nfs_mount_data *data,
																 struct nfs_cache **d_cache)
{
	struct nfs_cache *nfs_cache_current = nfs_cache_first;
	int i = 0;
	while (nfs_cache_current != NULL) {
		if (++i > 128) {
			/* 128 is just arbitrarily big */
			printk("NFS: nfs_create_new_cache_if_no_reuse, looping\n");
			break;
		}
		if (!strncmp(nfs_cache_current->root,data->disk_cache_root,NFS_MAX_CACHE_ROOT_PATH)) {
			if (nfs_cache_current->min_free < data->disk_cache_min_free)
				nfs_cache_current->min_free = data->disk_cache_min_free;
			nfs_cache_current->cMountsUsing++;
			(*d_cache) = nfs_cache_current;
			return 1;
		}
		nfs_cache_current = nfs_cache_current->next;
	}
	/* did not find it, so create one */
	(*d_cache) = (struct nfs_cache *) kmalloc(sizeof(struct nfs_cache),GFP_KERNEL);
	verb_nfs_printk("NFS: nfs_create_new_cache_if_no_reuse, kmalloc-ed at %lx\n",
									(long unsigned) *d_cache);
	/* insert at head of the list */
	(*d_cache)->next = nfs_cache_first;
	nfs_cache_first = (*d_cache);
	return 0;
}

int /* error */
nfs_set_cache_dir_and_fullstate(struct nfs_server *server,struct nfs_mount_data *data,
																struct nfs_sb_info *nfs_sb)
{	
	struct nfs_cache *cache;
	int status;
	verb_nfs_printk("NFS: nfs_set_cache_dir_and_fullstate: called\n");
	/* first set the max size, since that gets set even if we reuse
		 a cache */
	nfs_sb->dc_max_size = data->disk_cache_max_size;
	nfs_sb->dc_pages_used = 0;
	if (nfs_create_new_cache_if_no_reuse(data,&nfs_sb->s_cache)) {
		/* we are reusing a cache and have just set nfs_sb->s_cache to point to it,
			 so we're done */
		verb_nfs_printk("NFS: nfs_set_cache_dir_and_fullstate: reusing\n");
		return 0;
	}
	/* otherwise, nfs_sb->s_cache now points to the struct nfs_cache that we should fill in */
	cache = nfs_sb->s_cache;
	if (!cache) {
		printk("nfs_set_cache_dir_and_fullstate: got a null d_cache\n");
		return 1; /* error */
	}
	strncpy(cache->root,data->disk_cache_root,NFS_MAX_CACHE_ROOT_PATH);
	cache->min_free = data->disk_cache_min_free;
	cache->cUpdatesBeforeNextStat = NFS_UPDATES_BETWEEN_STATS;
	cache->cMountsUsing = 1;

	verb_nfs_printk("NFS: nfs_set_cache_dir_and_fullstate, getting directory inode for %s\n",
									cache->root);
	{
		uid_t old_fsuid = current->fsuid;
		current->fsuid = 0; /* be root */
		status = open_namei(cache->root,O_RDONLY,0,&cache->inodeDir,NULL);
		current->fsuid = old_fsuid;
	}
	if (status) {
		if (status == EACCES || status == -EACCES)
			printk("NFS: nfs_set_cache_dir_and_fullstate: could not open cache directory, Permission denied\n");
		else 
			printk("NFS: nfs_set_cache_dir_and_fullstate: could not open cache directory = \"%s\"; status = %d\n",
						 cache->root,status);
		server->flags &= ~NFS_MOUNT_DC_ENABLE;
		cache->inodeDir = NULL;
		return 1;  /* error */
	}
	nfs_printk("NFS: nfs_set_cache_dir_and_fullstate, getting full-state inode for %s\n",
						 cache->root);
	/* open_namei decrements the count of the directory, so increment it first */
	cache->inodeDir->i_count++;
	/* note the leading dot in the below file name */
	{
		uid_t old_fsuid = current->fsuid;
		current->fsuid = 0; /* be root */
		status = open_namei(".FULL",O_CREAT|O_APPEND,NFS_CACHE_STATUS_MODE,
												&cache->inodeFullStateIndicator,
												cache->inodeDir);
		current->fsuid = old_fsuid;
	}
	if (status) {
		printk("NFS: nfs_set_cache_dir_and_fullstate: could not open cache full state file; status = %d\n",
					 status);
		server->flags &= ~NFS_MOUNT_DC_ENABLE;
		cache->inodeFullStateIndicator = NULL;
		iput(cache->inodeDir);
		cache->inodeDir = NULL;
		return 1; /* error */
	} 
	cache->cUpdatesBeforeNextStat = NFS_UPDATES_BETWEEN_STATS;
	return 0;
}

static
void
nfs_remove_cache_from_list(struct nfs_cache *cache)
{
	struct nfs_cache *nfs_cache_current = nfs_cache_first;
	struct nfs_cache *nfs_cache_prior = NULL;
	while (nfs_cache_current != cache && nfs_cache_current != NULL) {
		nfs_cache_prior = nfs_cache_current;
		nfs_cache_current = nfs_cache_current->next;
	}
	if (nfs_cache_current == NULL) {
		printk("NFS: nfs_remove_cache_from_list: could not find cache %s\n",
					 cache->root);
		return;
	}
	if (nfs_cache_current == nfs_cache_first) {
		nfs_cache_first = nfs_cache_current->next;
		verb_nfs_printk("NFS: nfs_remove_cache_from_list: kfree after removing from head\n");
		kfree(nfs_cache_current);
		nfs_cache_current = NULL;
		return;
	}
	/* skip over that item */
	nfs_cache_prior->next=nfs_cache_current->next;
	kfree(nfs_cache_current);
	nfs_cache_current = NULL;
	verb_nfs_printk("NFS: nfs_remove_cache_from_list: kfree after removing from middle/end\n");
}
	
void
nfs_unuse_cache(struct super_block *sb) 
{
	struct nfs_cache *cache = sb->u.nfs_sb.s_cache;
	if 	(--cache->cMountsUsing == 0) {
		verb_nfs_printk("NFS: nfs_unuse_cache: putting cache dir inode for dev %d\n",sb->s_dev);
		if (cache->inodeDir->i_count != 0) {
			iput(cache->inodeDir);
		} else {
			printk("NFS: nfs_unuse_cache, inodeDir already at 0 count\n");
		}
		verb_nfs_printk("NFS: putting full state inode for dev %d\n",sb->s_dev);
		if (cache->inodeFullStateIndicator->i_count != 0) {
			iput(cache->inodeFullStateIndicator);
		} else {
			printk("NFS: nfs_unuse_cache, inodeFullStateIndicator already at 0 count\n");
		}
		nfs_remove_cache_from_list(cache);
	}
}


static inline
int
nfs_make_cache_filename(char *sz, struct inode *inode)
{
	char *szHostname = NFS_HOSTNAME(inode);
#ifdef NFS_USE_SHORTER_CACHEFILE_NAME
	int cch = strlen(szHostname);
	/* server, device, inode used as unique description of file */
	/* use a comma to separate the host name from the device-inode */
	return sprintf(sz,"%3s%2s,%ld-%lu",szHostname,
								 szHostname+cch-2-1,inode->i_dev,inode->i_ino);
#else
  /* FIXGJB: is 7 enough?  32 bits unsigned needs 10 digits for 4294967296 */
	return sprintf(sz,"%s,%02ld-%07lu",szHostname,inode->i_dev,inode->i_ino);
#endif
}

static inline
int
nfs_make_cache_filename_server_dev_num(char *sz, struct nfs_server *server, int dev, long unsigned num)
{
	/* server, device, inode used as unique description of file */
	return sprintf(sz,"%s-%02ld-%07lu",server->hostname,dev,num);
}



int /* error */
nfs_unlink_cache_file(struct inode *inodeCache)
{
	int status = 0;
	nfs_printk("NFS: nfs_unlink_cache_file called\n");

	if (inodeCache && inodeCache->i_ibcinfo) {
		char basename[NFS_CACHEFILE_MAXLEN];
		int namelen = 
			nfs_make_cache_filename_server_dev_num(basename,
																						 &(inodeCache->i_ibcinfo->nfs_sb_info->s_server),
																						 inodeCache->i_ibcinfo->device,
																						 inodeCache->i_ibcinfo->inode_number);
		struct inode_being_cached_info *pbci = inodeCache->i_ibcinfo;
		struct inode *inodeCacheDir = NULL;
		char *pchBitmap = NULL;
		struct nfs_sb_info *nfs_sbi = NULL;
		int cpagesValid = 0;
		if (pbci == NULL) {
			printk("NFS: nfs_unlink_cache_file, pbci = NULL local (%d,%ld)\n",
						 inodeCache->i_dev,inodeCache->i_ino);
			return 1; /* error */
		}
		if (pbci->fInodeNFSInMemory == FALSE) {
			nfs_printk("NFS: nfs_unlink_cache_file, (%d,%ld) fIsInMemory == false\n",
								 pbci->device,pbci->inode_number);
		}
		/* make sure it's not marked as complete */
		inodeCache->i_mode &= ~NFS_COMPLETED_BIT;
		pchBitmap = pbci->pchValidBitmap;
		inodeCacheDir = pbci->inodeDir;
		cpagesValid = pbci->cpagesValid;
		nfs_sbi = pbci->nfs_sb_info;
		if (pchBitmap) {
			inodeCache->i_ibcinfo->pchValidBitmap = NULL;
			kfree(pchBitmap);
			nfs_printk("NFS: nfs_unlink_cache_file: %s kfree-d bitmap\n",basename);
		} else {
			/* we should not be removing something that doesn't have a bitmap
				 since that would imply that it's complete, and needn't be removed
				 by the kernel --gjb */
			printk("NFS: nfs_unlink_cache_file: %s bitmap was null\n",basename);
		}
		nfs_printk("NFS: nfs_unlink_cache_file, testing fInodeNFSInMemory\n");
		if (nfs_sbi == NULL) {
			printk("NFS: nfs_unlink_cache_file, (%d,%ld)->(%d,%ld)\n",
						 pbci->device,pbci->inode_number,
						 inodeCache->i_dev,inodeCache->i_ino);
		} else {
			nfs_printk("NFS: nfs_unlink_cache_file, updating pages used, was = %d, sub %d\n",
								 nfs_sbi->dc_pages_used,cpagesValid);
			if (nfs_sbi->dc_pages_used > cpagesValid)
				nfs_sbi->dc_pages_used -= cpagesValid;
			else {
				nfs_sbi->dc_pages_used = 0;
				printk("NFS: nfs_unlink_cache_file, removing more space than was used, %d removed from %d\n",
							 cpagesValid, nfs_sbi->dc_pages_used);
				}
		}
		nfs_printk("NFS: checking inodeCacheDir\n");
		inodeCacheDir->i_count++;
		if (inodeCacheDir && inodeCacheDir->i_op && inodeCacheDir->i_op->unlink) {
			nfs_printk("NFS: nfs_unlink_cache_file, performing on %s, len %d, dir = (%d,%ld)\n",
								 basename,namelen,inodeCacheDir->i_dev,inodeCacheDir->i_ino);
			status = inodeCacheDir->i_op->unlink(inodeCacheDir,basename,namelen);
		} else {
			printk("NFS: nfs_unlink_cache_file, cannot unlink, cache dir == NULL, (%d,%ld)\n",
						 inodeCache->i_dev,inodeCache->i_ino);
		}
	}
	nfs_printk("NFS: nfs_unlink_cache_file, returning\n");
	return status;
}

void
nfs_clear_inode_hook(struct inode *inode)
{
	/* do not put debug code here-- it happens so fast that
		 it might have been overflowing the syslog buffer
		 in some wierd way --gjb */
	struct inode * inodeCache = NFS_CACHE_INODE(inode);
	struct inode_being_cached_info *pbci;
	inode->clear_inode_hook = NULL;
	verb_nfs_printk("NFS: nfs_clear_inode_hook (%d,%ld) -> (%d,%ld)\n",
			inode->i_dev,inode->i_ino,
			inodeCache?inodeCache->i_dev:-1,
			inodeCache?inodeCache->i_ino:-1);
	if (inodeCache) {
		/* we want to iput the inodeCahe if it's not really got much of it read,
			 since it wastes an inode spot, and we're not gonna bother filling
			 it in */
		if (nfs_is_mostly_uncached_file(inodeCache)) {
			/* it will get cleaned up when clear_inodeCacheHook gets called */
			if (inodeCache->i_count > 0)
				iput(inodeCache);
		} 
		/* iput the cache inode if there is one and if it doesn't need filling in
			 if it does need feeling in, we iput it when it's filled in completely --gjb */
		if (nfs_is_completely_cached_file(inodeCache)) {
			if (inodeCache->i_count > 0)
				iput(inodeCache);
		}
		pbci = inodeCache->i_ibcinfo;
		if (pbci && pbci->fInodeNFSInMemory) {
			memcpy(&pbci->u.fhandle,&inode->u.nfs_i.fhandle,sizeof(struct nfs_fh));
			pbci->u.fhandle = inode->u.nfs_i.fhandle;
			pbci->fInodeNFSInMemory = FALSE;
			pbci->u.inodeNFS = NULL;
		}
	} else {
		/* should not call this if we do not have a inodeCache */
		printk("NFS: nfs_clear_inode_hook (%d,%ld) has no inodeCache\n",
					 inode->i_dev,inode->i_ino);
	}
}

void
nfs_clear_inodeCache_hook(struct inode *inodeCache)
{
	struct inode_being_cached_info *pbci = inodeCache->i_ibcinfo;
	struct inode *inode = NULL;

	inodeCache->clear_inode_hook = NULL;

	nfs_printk("NFS: nfs_clear_inodeCache_hook, called\n");

	/* this should not happen...  I think the inodeCache will hang around
		 until we iput it, and that only happens when it is completed. 
		 it'd be nice to not have to deal with unlinking cached files, 
		 if possible --gjb */
	if (pbci == NULL) {
		printk("NFS: nfs_clear_inodeCache_hook, inodeCache->ibcinfo == NULL, local (%d,%ld)\n",
					 inodeCache->i_dev, inodeCache->i_ino);
		return;
	}
	if (!nfs_is_completely_cached_file(inodeCache)) {
			nfs_printk("NFS: nfs_clear_inodeCache_hook:  (%d,%ld) -> (%d,%ld) being removed\n",
											pbci->device, pbci->inode_number, inodeCache->i_dev,inodeCache->i_ino);
			/* now remove it */
			if (nfs_unlink_cache_file(inodeCache)) {
				printk("NFS: nfs_clear_inodeCache_hook, error removing cache file (%d,%ld)->(%d,%ld)\n",
							 pbci->device,pbci->inode_number,
							 inodeCache->i_dev,inodeCache->i_ino);
			}
	}

	/* now we need to make sure the inode that it's caching
		 doesn't point to us any more and we free the structures 
		 hanging off of it */
	if (pbci->fInodeNFSInMemory) {
		inode = pbci->u.inodeNFS;
		nfs_printk("NFS: nfs_clear_inodeCache_hook, (%d,%ld)-> (%d,%ld), NFS inode in memory\n",
					 inode->i_dev,inode->i_ino,
					 inodeCache->i_dev,inodeCache->i_ino);
		if (inode == NULL) {
			printk("NFS: nfs_clear_inodeCache_hook, inode == NULL, but fInodeNFSInMemory (%d,%ld)\n",
						 inodeCache->i_dev,inodeCache->i_ino);
		} else {
			if (NFS_CACHE_INODE(inode) == NULL) {
				printk("NFS: nfs_clear_inodeCache_hook, (%d,%ld)->(%d,%ld), inode's inodeCache is already null\n",
							 inode->i_dev,inode->i_ino,
							 inodeCache->i_dev,inodeCache->i_ino);
			}
			/* unpoint it to us, so it redoes the open */
			NFS_CACHE_INODE(inode) = NULL;
		}
	}
	inodeCache->i_ibcinfo = NULL;
	kfree(pbci);
	nfs_printk("NFS: nfs_clear_inodeCache_hook, done. kfree-d inodeCache->i_ibcinfo (%d,%ld)\n",
						 inodeCache->i_dev,inodeCache->i_ino);
}

static inline
char *
nfs_kmalloc_bitmap(int num_pages)
{
	/* >> 3 = divide by 8; kernel compiled w/ no-strength-reduce */
  int num_bytes = ((num_pages - 1) >> 3) + 1;
  char *pch = (char *) kmalloc(num_bytes,GFP_KERNEL);
	if (pch == NULL) {
		printk("NFS: nfs_kmalloc_bitmap -- could not allocate %d bytes\n",num_bytes);
	}
	verb_nfs_printk("NFS: nfs_kmalloc_bitmap allocated %d bytes for %d pages\n",
								 num_bytes, num_pages);
	memset(pch,0,num_bytes);
  return pch;
}

static inline
void
nfs_initialize_bci(struct inode_being_cached_info *pbci, struct inode * inode,
									 struct inode *inodeCache)
{
  int num_pages = (((long) inode->i_size - 1) >> PAGE_SHIFT) + 1;
	verb_nfs_printk("NFS:nfs_initialize_bci, (%d,%ld), size = %lu, numpages=%d\n",
									inode->i_dev,inode->i_ino,
									inode->i_size,num_pages);
	pbci->u.inodeNFS = inode;
	pbci->fInodeNFSInMemory = TRUE;
	pbci->device = inode->i_dev;
	pbci->inode_number = inode->i_ino;
	pbci->nfs_sb_info = &(inode->i_sb->u.nfs_sb);
	pbci->inodeDir = inode->i_sb->u.nfs_sb.s_cache->inodeDir;
	if (inodeCache && nfs_is_completely_cached_file(inodeCache)) {
		nfs_printk("NFS: nfs_initialize_bci, avoiding kmalloc since file is complete\n");
		pbci->pchValidBitmap = NULL;
		pbci->cpagesValid = pbci->cpagesTotal = num_pages;
	}	else {
		pbci->pchValidBitmap = nfs_kmalloc_bitmap(num_pages);
		pbci->cpagesValid = 0;
		pbci->cpagesTotal = num_pages;
	}
}

/* must kfree pbci->pchValidBitmap after this call */
/* still must set inodeCache member, too */
static inline
struct inode_being_cached_info *
nfs_kmalloc_initialize_bci(struct inode * inode, struct inode *inodeCache)
{
	struct inode_being_cached_info *pbci;
	pbci = (struct inode_being_cached_info *)
		kmalloc(sizeof(struct inode_being_cached_info), GFP_KERNEL);
	nfs_initialize_bci(pbci,inode,inodeCache);
	return pbci;
}

static inline
void
nfs_mark_bitmap_valid_for_ipage(struct inode_being_cached_info *pbci, unsigned long ipage)
{
  int ibyte = ipage >> 3; /* /8 */
  int ibit = ipage - (ibyte << 3); /* *8 */
#ifndef NO_DEBUG_CACHE_BITMAP
	if (nfs_is_bitmap_valid_for_ipage(pbci->pchValidBitmap,ipage)) {
		printk("NFS: nfs_mark_bitmap_valid_for_ipage already valid (byte,bit) = (%d,%d)\n",
					 ibyte, ibit);
		pbci->cpagesValid--;
	}
#endif
  pbci->pchValidBitmap[ibyte] |= 1<<(7-ibit);
	pbci->cpagesValid++;
  verb_nfs_printk("NFS:nfs_mark_bitmap_valid_for_ipage %lu (byte,bit) = (%d,%d), now %d/%d\n",
									ipage, ibyte, ibit,
									pbci->cpagesValid,pbci->cpagesTotal);
}

static inline
int /* bool */
nfs_is_bci_complete(const struct inode_being_cached_info *const pbci)
{
	return pbci->cpagesValid == pbci->cpagesTotal;
}

/* The caller of this frees the bitmap */
static inline
void
nfs_mark_as_completely_cached_file_and_free_bitmap(struct inode *inode)
{
	struct inode *inodeCache = NFS_CACHE_INODE(inode);
	char *pchBitmap;
	if (inodeCache == NULL) {
		printk("NFS: nfs_mark_as_completely_cached_file_and_free_bitmap, (%d,%ld) inodeCache == NULL\n",
					 inode->i_dev,inode->i_ino);
		return;
	}
	if (inode->i_mtime != inodeCache->i_mtime) {
		printk("NFS: nfs_mark_as_completely_cached_file_and_free_bitmap, (%d,%ld)->(%d,%ld) mtime %ld vs %ld\n",
					 inode->i_dev,inode->i_ino,
					 inodeCache->i_dev,inodeCache->i_ino,
					 inode->i_mtime,inodeCache->i_mtime);
	}
					 
	verb_nfs_printk("NFS: nfs_mark_as_completely_cached_file_and_free_bitmap: (%d,%ld)->(%d,%ld)\n",
									inode->i_dev,inode->i_ino,
									inodeCache->i_dev,inodeCache->i_ino);
						 
	if (inodeCache->i_ibcinfo) {
		pchBitmap=inodeCache->i_ibcinfo->pchValidBitmap;
		inodeCache->i_ibcinfo->pchValidBitmap = NULL;
		kfree(pchBitmap);
	} else {
		printk("NFS: nfs_mark_as_completely_cached_file_and_free_bitmap, pbci == NULL\n");
	}
	inodeCache->i_mode |= NFS_COMPLETED_BIT;
	inodeCache->i_dirt = 1;
}


/* Set the NFS_COMPLETED_BIT on the .FULL file for this sb's cache directory
   and possible the NFS_ERROR_WRITING_BIT, if enumType==NFS_FILLED_DUE_TO_WRITE_ERROR
   Also output the CURRENT_TIME, follwowed by server,s_dev if it was 
	 a max space violation */
static inline
void nfs_mark_cache_dir_as_filled(struct super_block *sb, int enumType)
{
	if (sb->u.nfs_sb.s_cache) {
		struct inode *inode_fs = sb->u.nfs_sb.s_cache->inodeFullStateIndicator;
		struct file_operations *f_op = inode_fs->i_op->default_file_ops;
		struct file fileCache;
		char szTime[256+12];
		int cbToWrite;
		int fs = get_fs(), ds = get_ds();
		if (enumType == NFS_FILLED_DUE_MAX_SPACE_VIOL) {
			cbToWrite = sprintf(szTime,"%lu %s,%d\n",CURRENT_TIME,
													sb->u.nfs_sb.s_server.hostname,
													sb->s_dev);
		} else if (enumType == NFS_FILLED_DUE_TO_WRITE_ERROR) {
			cbToWrite = sprintf(szTime,"%lu WRITE ERROR\n",CURRENT_TIME);
		} else {
			/* if it's a min free violation, we do not need the server info */
			cbToWrite = sprintf(szTime,"%lu\n",CURRENT_TIME);
		}
		inode_fs->i_dirt = 1; /* make sure we write it */
		switch (enumType) {
		case NFS_FILLED_DUE_TO_WRITE_ERROR:
			inode_fs->i_mode |= NFS_ERROR_WRITING_BIT;
			inode_fs->i_mode |= NFS_MIN_FREE_VIOL_BIT;
			break;
		case NFS_FILLED_DUE_MIN_FREE_VIOL:
			inode_fs->i_mode |= NFS_MIN_FREE_VIOL_BIT;
			break;
		case NFS_FILLED_DUE_MAX_SPACE_VIOL:
			inode_fs->i_mode |= NFS_MAX_SPACE_VIOL_BIT;
			break;
		}
		
		fileCache.f_op = f_op;
		fileCache.f_mode = WRITEA;  /* 3 */
		fileCache.f_flags = 0;
		fileCache.f_count = 1;
		fileCache.f_inode = inode_fs;
		fileCache.f_pos = inode_fs->i_size; /* append */
		fileCache.f_reada = 0;
	
		/* now write the time to the file status indicator inode,
			 so the user level process gets awakened */
		if (f_op->write) {
			uid_t old_fsuid = current->fsuid;
			current->fsuid = 0; /* "become" root */
			set_fs(ds);
			f_op->write(inode_fs,&fileCache,szTime,cbToWrite);
			set_fs(fs);
			current->fsuid = old_fsuid;
		}
	} else {
		printk("NFS: nfs_mark_cache_dir_as_filled, sb->u.nfs_sb.s_cache == NULL\n");
	}
}
	


int /* error */
nfs_open_cache_file(struct inode *inode, int fNoCreate)
{ /* scope */
	int retval = 0;
	int fs = get_fs(), ds = get_ds();
	struct inode *inodeCache = NFS_CACHE_INODE(inode);
	time_t server_mtime = inode->i_mtime;

	if (inodeCache == 0 || inodeCache->i_ino == 0) {
		char szCachePath[NFS_CACHEFILE_MAXLEN];
		struct inode *inodeCacheDir = NFS_CACHE_DIR_INODE(inode);
		nfs_make_cache_filename(szCachePath,inode);
		/* inode->i_version seems to change -- do not use it
			 (is that the inode generation? they're big numbers.... where
			 is the generation # if that's not it? ) --gjb */
		/* open_namei decrements the directory count, but we want it to stay > 0 --gjb */
		inodeCacheDir->i_count++;
		{ /* careful about our memory segment */
			uid_t old_fsuid = current->fsuid;
			current->fsuid = 0; /* "Become" root */
			set_fs(ds);
			/* FIXGJB: should below not be O_WRONLY, 0 or O_RDWR instead? */
			retval = open_namei(szCachePath,(O_WRONLY|(fNoCreate?0:O_CREAT)),
													NFS_CACHE_MODE,&inodeCache,inodeCacheDir);
			set_fs(fs);
			current->fsuid = old_fsuid;
		}

		if (retval != 0) {
			if (!nfs_cache_dir_filled(inode)) {
				/* do not bother with the warning if the cache directory is full */
				printk("NFS: nfs_open_cache_file, open of cache file \"%s\" failed, retval = %d\n",
							 szCachePath, retval);
			}
			inodeCache = NULL;
			NFS_CACHE_INODE(inode) = NULL;
			return 1;
		}

#ifdef DEBUG_NFS
		if (inodeCache == NULL) {
			printk("NFS: nfs_open_cache_file, inodeCache == NULL!\n");
			return 1;
		}
#endif

		/* we want to the mtime of the cache file's inode 
			 to be the server's mtime,
			 then we use can use that mtime to check if the server's copy
			 of the file has changed */
		inodeCache->i_mtime = server_mtime;
		inodeCache->i_size = inode->i_size;
		inodeCache->i_uid = NFS_CACHE_UID;
		inodeCache->i_gid = NFS_CACHE_GID;
		inodeCache->i_dirt = 1;
#ifdef DEBUG_NFS
		if (inodeCache->i_ibcinfo) {
			struct inode_being_cached_info *pbci = inodeCache->i_ibcinfo;
			inodeCache->clear_inode_hook = NULL;
			inodeCache->i_ibcinfo = NULL;
			kfree(pbci);
			printk("NFS: nfs_open_cache_file, unexpected kfree-ing inodeCache->i_ibcinfo\n");
		}
#endif
		/* we pass inodeCache in, rather than swapping the order of the next
			 two lines, to avoid a race condition whereby the pbci is not intialized,
			 but an inode is pointing to the inodeCache already --gjb */
		NFS_CACHE_INODE(inode) = NULL;
		inodeCache->i_ibcinfo = nfs_kmalloc_initialize_bci(inode,inodeCache);
		NFS_CACHE_INODE(inode) = inodeCache;
		inode->clear_inode_hook = nfs_clear_inode_hook;
		inodeCache->clear_inode_hook = nfs_clear_inodeCache_hook;
		/* save this inode used for caching in the inode of the file we're caching */
		verb_nfs_printk("NFS: nfs_open_cache_file, inodeCache->i_ibcinfo = %lx\n",
										(unsigned long) inodeCache->i_ibcinfo);
	} 	else {
		/* just use the one we've already got -- we do not expect this to happen,
		 since we shouldn't have called open, then */
		printk("NFS: nfs_open_cache_file, inodeCache already exists: (%d,%ld) on local (%d,%ld)\n",
					 inode->i_dev, inode->i_ino,inodeCache->i_dev,inodeCache->i_ino);
	}
	return 0;
}

/* Note that the super block we pass in is that of the NFS system,
   not of the local caching system */
static inline
void
_nfs_check_too_full(struct super_block *sb_nfs)
{
	int bsize, blocks, bavail;
	int fs = get_fs(), ds = get_ds();
	struct statfs cache_statfs;
	int kb_left;
	struct super_block *sb = sb_nfs->u.nfs_sb.s_cache->inodeFullStateIndicator->i_sb;
	if (!sb || !sb->s_op || !sb->s_op->statfs) {
		printk("NFS: nfs_check_too_full, can not statfs filesystem\n");
		return;
	}
	set_fs(ds);
	sb->s_op->statfs(sb,&cache_statfs,sizeof(struct statfs));
	set_fs(fs);
	bsize = cache_statfs.f_bsize;
	blocks = cache_statfs.f_blocks;
	bavail = cache_statfs.f_bavail;
	verb_nfs_printk("NFS: nfs_check_too_full, free space statfs = %d x %d / %d\n",
						 bsize, bavail, blocks);
	sb_nfs->u.nfs_sb.s_cache->bavail = bavail;
	kb_left = bavail*bsize >> 10; /* divide by 1024 = 1kb */
	if (kb_left < sb_nfs->u.nfs_sb.s_cache->min_free) {
		printk("NFS: nfs_check_too_full, filled: left %d < reqd %d\n",
					 kb_left, sb_nfs->u.nfs_sb.s_cache->min_free);
		nfs_mark_cache_dir_as_filled(sb_nfs,NFS_FILLED_DUE_MIN_FREE_VIOL);
	}
}

/* This one is externally visible */
void nfs_check_too_full(struct super_block *sb) { _nfs_check_too_full(sb); }

void
nfs_invalidate_disk_cache(struct inode *inode, struct inode *inodeCache)
{
	char *pchBitmap;
	/* file has changed, so empty our valid pages bitmap */
	inodeCache->i_mode &= ~NFS_COMPLETED_BIT;
	inodeCache->i_dirt = 1;
	inodeCache->i_mtime = inode->i_mtime;
	if ((pchBitmap = inodeCache->i_ibcinfo->pchValidBitmap) != NULL) {
		nfs_printk("NFS: nfs_invalidate_disk_cache, kfree-ing (invalid)\n");
		inodeCache->i_ibcinfo->pchValidBitmap = NULL;
		kfree(pchBitmap);
	}
	if (inodeCache->i_ibcinfo == NULL) {
		printk("NFS: nfs_invalidate_disk_cache, inodeCache->i_ibcinfo == NULL, kmalloc-ing\n");
		inodeCache->i_ibcinfo = nfs_kmalloc_initialize_bci(inode,inodeCache);
	} else {
		DC_PAGES_USED_FROM_NFS(inode) -= inodeCache->i_ibcinfo->cpagesValid;
		nfs_initialize_bci(inodeCache->i_ibcinfo,inode,inodeCache);
	}
	nfs_printk("NFS: nfs_invalidate_disk_cache, inodeCache->i_ibcinfo = %lx\n",
						 (unsigned long) inodeCache->i_ibcinfo);
	/* Do not need to call nfs_unlink_cache_file; do that at clear inode
		 or put_super time, meanwhile, we can just write where we were writing,
		 and let the kernel structures reflect their validity --gjb */
}

/* Write the just read data to our local cache,
	 update our in-memory data structures for what is valid for that cache file
	 This is called from at least
	 nfs_readpage_sync and nfs_readpage_result (latter is for async calls)
	 -- gjb
	 */
void
nfs_update_read_cache(struct inode *inode,
											unsigned long offset,
											char *buffer,
											unsigned int countAsked,
											int result,
											int fUserSpace)
{
	unsigned long filesize = inode->i_size;
	struct inode_being_cached_info *pbci;
	struct nfs_cache *s_cache;
  /* check the caching filesystem's free space every 256 cache updates;
		 it's not real expensive, but we'd like to avoid a fn call */
	time_t server_mtime = inode->i_mtime;

	/* result is the # bytes read as returned by _result, or _sync,
		 countAsked is how many bytes we asked for
		 We compute how many bytes
		 to write so tha we don't finish off the last page w/ nulls --gjb */
	unsigned int cbToWrite = offset + result > filesize? 
		filesize-offset /* needs truncation, so to speak */ 
		: result;
	int cbytes = 0;
	int retval = 0;
	int fs = get_fs(), ds = get_ds();
	struct inode *inodeCache = NFS_CACHE_INODE(inode);
	struct file_operations *f_op = NULL;
	struct file fileCache;
	extern int errno;

	/* ensure that we have space */
	if (nfs_cache_dir_filled(inode)) {
		verb_nfs_printk("NFS: nfs_update_read_cache, cache full (min_free) not updating\n");
		return;
	}

	/* and ensure that we do not exceed our allowed space for this mounted parititon */
	if (DC_PAGES_USED_FROM_NFS(inode) >= DC_MAX_SIZE_FROM_NFS(inode)) {
		verb_nfs_printk("NFS: nfs_update_read_cache, max_size exceeded, not updating\n");
		return;
	}

	/* Now make sure that inodeCache is valid */
	if (inodeCache == 0 || inodeCache->i_ino == 0) {
		if (nfs_open_cache_file(inode,0)) {
			/* failed */
			printk("NFS: nfs_update_read_cache, open failed!\n");
			return;
		}
		inodeCache = NFS_CACHE_INODE(inode);
	}

	/* I think we should always have the counts non-zero --gjb */
	if (inode->i_count == 0) {
		printk("NFS: nfs_update_read_cache, inode->i_count == 0\n");
	}
	if (inodeCache->i_count == 0) {
		printk("NFS: nfs_update_read_cache, inodeCache->i_count == 0\n");
	}

	/* if the mtime or size has changed */
	if (server_mtime != inodeCache->i_mtime || filesize != inodeCache->i_size) {
		struct inode *inodeCache = NFS_CACHE_INODE(inode);
		nfs_printk("NFS: nfs_update_read_cache, inode has changed (%d,%ld)\nNFS:... server_mtime = %ld, inodeCache->i_mtime = %ld; size: %ld vs. %ld\n",
							 inode->i_dev,inode->i_ino,
							 server_mtime,inodeCache->i_mtime,
							 filesize,inodeCache->i_size);
		nfs_invalidate_disk_cache(inode,inodeCache);
	}

	/* we are done if the file is already completely cached, 
		 but we should not be here to begin with */
	if (nfs_is_completely_cached_file(inodeCache)) {
		printk("NFS: nfs_update_read_cache called on completely cached file\n");
		return;
	}

	s_cache = NFS_CACHE_INFO(inode);
#if NFS_UPDATES_BETWEEN_STATS != 0
	/* now the inodeCache is valid */
	if (--s_cache->cUpdatesBeforeNextStat <= 0) {
		s_cache->cUpdatesBeforeNextStat = NFS_UPDATES_BETWEEN_STATS;
		_nfs_check_too_full(inode->i_sb);
	}
#endif

	f_op = inodeCache->i_op->default_file_ops;
	pbci = inodeCache->i_ibcinfo;
	if (pbci == NULL) {
		printk("NFS: nfs_update_read_cache, pbci == NULL!\n");
		return;
	}

	fileCache.f_op = f_op;
	fileCache.f_mode = WRITEA;  /* 3 */
	fileCache.f_flags = 0;
	fileCache.f_count = 1;
	fileCache.f_inode = inodeCache;
	fileCache.f_pos = offset;
	fileCache.f_reada = 0;

	/* a bit counterintuitive-- the VFS operations are often null
		 for the common case -- ext2fs in particular has these both null  --gjb */
	/* FIXGJB: for efficiency, we could make these tests at mount
		 time and not permit that backend FS */
	if (f_op->open || f_op->lseek) {
		printk("NFS: nfs_update_read_cache -- Surprise! caching filesystem has open/lseek methods\n");
		if (f_op->open) {
			retval = f_op->open(inodeCache,&fileCache);
			verb_nfs_printk("NFS: open returned %d\n",retval);
		}
		if (retval !=0 ) {
			printk("NFS: nfs_update_read_cache, open returned %d\n",retval);
			return;
		}
		if (f_op->lseek) {
			retval = f_op->lseek(inodeCache, &fileCache, offset, 0);
		}
		if (retval !=0 ) {
			printk("NFS: nfs_update_read_cache, lseek returned %d\n",retval);
			return;
		}
	}
	/* there are no updatepage()/writepage() methods for ext2fs --
		 updatepage() is what generic file write uses --gjb */
	if (!f_op->write) {
		printk("NFS: nfs_update_read_cache -- f_op->write is NULL! (%d,%ld), pbci = %lx\n",
					 inodeCache->i_dev,inodeCache->i_ino,(unsigned long)inodeCache->i_ibcinfo);
		return;
	}

	if (fUserSpace)
		set_fs(ds);
	cbytes = f_op->write(inodeCache,&fileCache,buffer,cbToWrite);
	if (fUserSpace)
		set_fs(fs);

	if (cbytes != cbToWrite) {
		printk("NFS: nfs_update_read_cache: only wrote %d/%d bytes, errno = %d\n",
					 cbytes,cbToWrite,errno); 
		nfs_mark_cache_dir_as_filled(inode->i_sb,NFS_FILLED_DUE_TO_WRITE_ERROR);
	}
	inodeCache->i_dirt = 1; /* make sure we write it */
	inodeCache->i_mtime = server_mtime;
	verb_nfs_printk("NFS: nfs_update_read_cache: Writing %d/%d bytes @ %ld from (%d,%ld) to local (%d,%ld), now %d / %d\n",
									cbToWrite,countAsked,offset,
									inode->i_dev,inode->i_ino,
									inodeCache->i_dev,inodeCache->i_ino,
									pbci->cpagesValid,
									pbci->cpagesTotal);
		
	nfs_mark_bitmap_valid_for_ipage(inodeCache->i_ibcinfo,offset >> PAGE_SHIFT);
	++DC_PAGES_USED_FROM_NFS(inode);
	if (DC_PAGES_USED_FROM_NFS(inode) >= DC_MAX_SIZE_FROM_NFS(inode) - 1) {
		printk("NFS: nfs_update_read_cache, reached maximum pages cached allowed for %s,%d\n",
					 NFS_HOSTNAME(inode),inode->i_sb->s_dev);
		/* does not disallow caching altogether, only caching for
			 this mount; does not set the +x bit */
		nfs_mark_cache_dir_as_filled(inode->i_sb,NFS_FILLED_DUE_MAX_SPACE_VIOL);
	}

	if (nfs_is_bci_complete(pbci)) {
		nfs_mark_as_completely_cached_file_and_free_bitmap(inode);
	}
}


int /* bool */
nfs_cache_valid_for(struct inode *inode, struct page *page)
{
	struct inode *inodeCache = NFS_CACHE_INODE(inode);
	int fDoMessage = (page->offset == 0);

	verb_nfs_printk("NFS: nfs_cache_valid_for %lu?\n",page->offset >> PAGE_SHIFT);
	
	if (inodeCache == NULL || inodeCache->i_ino == 0) {
		int fForbidCreate = 0;
		/* we open, but we do not want to create if 
			 the cache dir is too full -- normally we do want to
			 create since otherwise we have to redo the open
			 once we do the update */
		if (nfs_cache_dir_filled(inode)) {
			NFS_INVALID_MSG("cache dir is filled");
			fForbidCreate = 1;
		}

		/* and ensure that we do not create if we're at our allowed space
			 for this mounted parititon already */
		if (DC_PAGES_USED_FROM_NFS(inode) >= DC_MAX_SIZE_FROM_NFS(inode)) {
			NFS_INVALID_MSG("overflowed max allowed for mount");
			fForbidCreate = 1;
		}

		if (nfs_open_cache_file(inode,fForbidCreate)) {
			NFS_INVALID_MSG("failed open");
			/* failed */
			return 0;
		}

		inodeCache = NFS_CACHE_INODE(inode);
		verb_nfs_printk("NFS: nfs_cache_valid_for, inodeCache->i_ibcinfo = %lx\n",
										(unsigned long) inodeCache->i_ibcinfo);
		if (inodeCache->i_ibcinfo == NULL) {
			printk("NFS: nfs_cache_valid_for, i_ibcinfo == NULL, (%d,%ld)->(%d,%ld)\n",
						 inode->i_dev,inode->i_ino,inodeCache->i_dev,inodeCache->i_ino);
			return 0;
		}
	}

	/* First check if the cache is no longer valid */
	if (inode->i_mtime != inodeCache->i_mtime) {
		verb_nfs_printk("NFS: invalid local cache -- mtime %ld vs %ld\n",	
										inode->i_mtime,inodeCache->i_mtime);
		NFS_INVALID_MSG("mtime has changed");
		return 0;
	}

	/* now check if the page is in our cache */
	if (nfs_is_completely_cached_file(inodeCache)) {
		verb_nfs_printk("NFS: nfs_cache_valid_for: HIT on %ld! (Whole file complete)\n",
							 page->offset >> PAGE_SHIFT);
		return 1;
	}
	if (nfs_is_bitmap_valid_for_ipage(inodeCache->i_ibcinfo->pchValidBitmap,
					  page->offset >> PAGE_SHIFT)) {
		verb_nfs_printk("NFS: nfs_cache_valid_for: HIT on %ld! (On given page)\n",
							 page->offset >> PAGE_SHIFT);
		return 1;
	}

	/* do not use the cache */
	NFS_INVALID_MSG("misc reason");
	return 0; /* false */
}

int /* error */
nfs_cache_readpage(struct inode *inode, struct page *page)
{
	char	*buffer = (char *) page_address(page);
	struct inode *inodeCache = NFS_CACHE_INODE(inode);
	struct file_operations *f_op = inodeCache->i_op->default_file_ops;
	int error = -1;
	int retval = 0;
	clear_bit(PG_error, &page->flags);

	/* this test should always fail, since nfs_cache_valid_for has the
		 same guard before returning true */
	if (!(inodeCache && inodeCache->i_ino != 0)) {
		printk("NFS: nfs_cache_readpage inodeCache not really valid\n");
		return -EIO;
	}

	verb_nfs_printk("NFS: nfs_cache_readpage from local i_ino = %ld\n",inodeCache->i_ino);
	if (f_op->open || f_op->lseek) {
		struct file fileCache;
		printk("NFS: nfs_cache_readpage -- Surprise! caching filesystem has open/lseek methods\n");
		if (f_op->open) {
			retval = f_op->open(inodeCache,&fileCache);
			nfs_printk("NFS: open returned %d",retval);
		}
		if (retval !=0 ) {
			printk("NFS: nfs_cache_readpage, open returned %d\n",retval);
		}
		retval = 0;
		if (f_op->lseek) {
			retval = f_op->lseek(inodeCache, &fileCache, page->offset, 0);
		}
		if (retval !=0 ) {
			printk("NFS: nfs_cache_readpage, open returned %d\n",retval);
		}
	}
	verb_nfs_printk("NFS: nfs_cache_readpage doing f_op->read\n");
	if (inodeCache->i_op->readpage) {
		/* This does the read asynchronously, but then
			 readpage is responsible for waking up the page, and setting
			 it's valid bit -- a lot like what nfs_readpage_async does,
			 relying on nfs_readpage_result to set the page up to date
			 and wake up the process --gjb */
		uid_t old_fsuid = current->fsuid;
		current->fsuid = 0; /* "become" root */
		error = inodeCache->i_op->readpage(inodeCache, page);
		current->fsuid = old_fsuid;
		verb_nfs_printk("NFS: i_op->readpage returned %d\n",error);
	}
	if (error) {
		/* either didn't read, or readpage returned an error */
		memset(buffer, 0, PAGE_SIZE);
		printk("NFS: nfs_cache_readpage returning ERROR\n");
		retval = -EIO;
	} else {
		/* success, so mark the cache used at this time */
		inodeCache->i_atime = CURRENT_TIME;
		inodeCache->i_dirt = 1;
		/* we need to update the i_atime on the server, too, in terms
			 of its CURRENT_TIME; i.e. inode->i_atime = CURRENT_TIME is not
			 good enough due to time skew problems */
		if (inodeCache->i_atime > inode->i_atime + NFS_CACHE_ATIME_IGNORE_SECONDS) {
			struct nfs_sattr sattr = {inode->i_mode, inode->i_uid, inode->i_gid, 
																inode->i_size, 
																{inodeCache->i_atime, 0}, 
																{inode->i_mtime, 0}  };
			struct nfs_fattr fattr;
			
			error = nfs_proc_setattr(NFS_SERVER(inode), NFS_FH(inode),
															 &sattr, &fattr);
		}
	}
	return retval;
}

struct super_block *
nfs_find_sb(char *hostname,kdev_t dev)
{
	struct super_block *sb;
	struct nfs_sb_info *nfs_sbi; 
	struct nfs_cache *nfs_cache;
	struct nfs_server *nfss;
	for (sb = super_blocks + 0 ; sb < super_blocks + NR_SUPER ; sb++) {		
		if (!sb->s_dev || sb->s_dev != dev)
			continue;
		if (sb->s_magic != NFS_SUPER_MAGIC)
			continue;
		nfs_sbi = &sb->u.nfs_sb;
		nfs_cache = nfs_sbi->s_cache;
		nfss = &nfs_sbi->s_server;
		if (strcmp(nfss->hostname,hostname)) /* if they do not match */
			continue;
		return sb;
	}
	return NULL;
}


int
nfs_read_space(struct inode *inode, struct file *file, char *buf, int count)
{
	int c;
	char szBuffer[128];
	if (verify_area(VERIFY_WRITE,buf,count) == -EFAULT) return -EFAULT;
	if (count < 16) return -EFAULT;
	if (file->f_pos > 0)
		return 0; /* pretend eof */
	c = sprintf(szBuffer,"Write format:\nServerName,device#,+/-cPages\ne.g.holden,2,-210\n");
	if (c>=count) return -EFAULT;
	memcpy_tofs(buf,szBuffer,c);
	file->f_pos++;
	return c;
}

/* from drivers/isdn/teles/callc.c	--gjb */
static int
my_atoi(char *s)
{
	int i, n;
	
	n = 0;
	if (!s)
		return -1;
	for (i = 0; *s >= '0' && *s <= '9'; i++, s++)
		n = 10 * n + (*s - '0');
	return n;
}


/* the write to /dev/nfs-cache-space procedure
	 parse what was written, and update appropriately
	 --gjb
	 */ 
int
nfs_update_space(struct inode *inode, struct file *file, const char *buf, int count)
{
	char szHostname[256+4+10];
	char *pch = szHostname;
	struct super_block *sb;
	int delta_space = 0;
	int i = 0;
	char chSign;
	kdev_t dev = 0;
	memcpy_fromfs(szHostname,buf,count);
	/* parse the string we got written to us */
	while (*pch != ',') {
		if (++i >= count)
			return -EFAULT;
		pch++;
	}
	*pch = '\0';
	dev = my_atoi(pch+1);
	pch+=3; i+= 3;
	if (dev>10) { pch++; i++; }
	if (i+1 >= count)
		return -EFAULT;
	chSign = *pch;
	pch++; i++;
	/* FIXGJB, consider '=' for setting absolute */
	delta_space = my_atoi(pch) * (chSign == '-'?-1:1);
	nfs_printk("NFS: nfs_update_space, %s, dev %d, delta_space=%d, chSign = %c\n",
						 szHostname,dev,delta_space,chSign);
	sb = nfs_find_sb(szHostname,dev);
	if (!sb) {
		printk("NFS: nfs_update_space, could not find superblock for %s dev %d",
					 szHostname,dev);
		return -EFAULT;
	}
	/* be sure to never let it go below zero */
	if (delta_space >= 0 ||
			-delta_space < sb->u.nfs_sb.dc_pages_used)
		sb->u.nfs_sb.dc_pages_used += delta_space;
	else {
		printk("NFS: nfs_update_space, too much space freed = %d, only %d used\n",
					 delta_space, sb->u.nfs_sb.dc_pages_used);
		sb->u.nfs_sb.dc_pages_used = 0;
	}
	return count;
}

/* cleanup inodes for the given device -- this gets done at unmount time,
	 and module cleanup time */
/* use dev = -1 to cleanup all devices, fUnlink = 0 to not unlink files */
/* Note: kdev_t is unsigned, so we use int dev, below, instead */
int
nfs_cleanup_inodes_with_partial_caches(int dev, int fUnlink) 
{
	struct inode *inode_first = get_first_inode();
	struct inode *inodeCache;
	struct inode_being_cached_info *pbci = NULL;
	int i = 0;
	int cRemoved = 0;
	int status = 0;
	nfs_printk("NFS: nfs_cleanup_inodes_with_partial_caches\n");
	/* loop a lot like we do for /proc/net/nfs-usage */
	for(inodeCache = inode_first; ; i++, inodeCache = inodeCache->i_next) {
		/* be sure we do not loop forever! FIXGJB, we'd like the
		 actually # of inodes nr_inodes, but that's not exported now */
		if (i>NR_INODE*8) {
			printk("NFS: nfs_cleanup_inodes_with_partial_caches, endless loop\n");
			break;
		}

		if (inodeCache == NULL)
			break;
		if (inodeCache == inode_first && i>0)
			break;
		if ((pbci = inodeCache->i_ibcinfo) == NULL)
			continue;
		/* we've got an inodeCache, check if it's from the device we're cleaning up for */
		if (dev != -1 && pbci->device != dev)
			continue;
		/* we have an inodeCache whose inode is invalidated due to unmount */
		if (fUnlink && (status = nfs_unlink_cache_file(inodeCache))) {
			printk("NFS: nfs_cleanup_inodes_with_partial_caches, error removing cache file (%d,%ld)->(%d,%ld) = %d\n",
						 pbci->device,pbci->inode_number,
						 inodeCache->i_dev,inodeCache->i_ino,status);
		}
		if (pbci->fInodeNFSInMemory) {
			printk("NFS: nfs_cleanup_inodes_with_partial_caches, found inode still in mem, (%d,%ld)->(%d,%ld), inodeNFS = %lx\n",
						 pbci->device,pbci->inode_number,
						 inodeCache->i_dev,inodeCache->i_ino,
						 (unsigned long) pbci->u.inodeNFS);
		}
			
		cRemoved++;
		inodeCache->clear_inode_hook = NULL;
		inodeCache->i_ibcinfo = NULL;
		kfree(pbci);

	}
	nfs_printk("NFS: nfs_cleanup_inodes_with_partial_caches, removed %d\n",
						 cRemoved);
	return cRemoved;
}
